/* -- updatedd: dyndns.c --
 *
 * Copyright (c) 2002, 2003, 2004 Philipp Benner <mail@philipp-benner.de>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * The GNU C Library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with the GNU C Library; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA.
 *
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#ifdef HAVE_GETOPT_LONG
# include <getopt.h>
#endif
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <base64encode.h>
#include <get_connection.h>
#include <error.h>
#include <version.h>
#ifndef HAVE_DPRINTF
# include <dprintf.h>
#endif
#include <utils.h>

#define DYNDNSHOST	"www.regfish.com"
#define PORT		80

#define MAXDATA 	512
#define	BUFSIZE		256

#define EXIT_OK		EXIT_CODE(0)
#define EXIT_WARN	EXIT_CODE(1)
#define EXIT_FATAL	EXIT_CODE(2)

void
  print_usage(char *argv, FILE *fp)
{
	(void)fprintf(fp,
		"\n                                Updatedd - \033[0;36;1mregfish\033[0m\n\n"
		"Usage: %s [-4 <ip>] [-6 <ip>] [-c] [-f] [-l <ttl>] [-m] -n <hostname> [-y]\n"
		"		(-u <e-mail>:<password> OR -t <token)\n\n"
		"	-h	--help			print this help\n"
		"	-4	--ipv4 <ip>		ipv4 address\n"
		"	-6	--ipv6 <ip>		ipv6 address\n"
		"	-c	--force-host		create hostname if it does not exist\n"
		"	-f	--force			force update\n"
		"	-L	--list-services		list supported services\n"
		"	-l	--ttl			time to live (300-86400)\n", argv);
	(void)fprintf(fp,
		" 	-m	--ismx			use this host as MX-host\n"
		"	-n	--hostname <hostname>	your hostname\n"
		"	-S	--service <service>	select the service\n"
		"	-t	--token <token>		security hash for login\n"
		"	-u	--user <user:pass>	user and password to login with\n"
		"	-v	--version		print version\n"
		"	-y	--with-syslog		print error messages to syslog\n\n");

	return;
}

struct arguments {
	char *hostname;
	char *ipv4;
	char *ipv6;
	char *user;
	char *token;
	char *ttl;
	int   ismx;
	int   force;
	int   forcehost;
};

int i_syslog = 0;

struct ip_addresses {
	struct in_addr real_ip;
	struct in_addr cached;
};

static struct return_codes {
	char *code;
	char *message;
	int  error;
} return_codes[] = {
	{ "401", "standard authentication failed: wrong username or password",	1 },
	{ "402", "secure authentication failed: wrong security hash",		1 },
	{ "403", "standard authentication failed: domain name does not exist",	1 },
	{ "404", "regfish is not authorized to change the domain",		1 },
	{ "405", "invalid SOA-zone, please contact regfish.com", 		1 },
	{ "406", "resource record does not exist, use force-host",		1 },
	{ "407", "invalid ttl value",						1 },
	{ "408", "invalid ipv4 address",					1 },
	{ "409", "invalid ipv6 address",					1 },
	{ "410", "invalid auth type, please send a bugreport to the author",	1 },
	{ "411", "could not update MX-host",					1 },
	{ "412", "invalid hostname",						1 },
	{ "413", "unsecure authentication is not admitted",			1 },
	{ "414", "unknown fault",						1 },
	{ "415", "multiple A-pointer found",					1 },
	{ "101", "no changes, update considered abusive",			0 },
	{ "100", "update good and successful, ip updated",			0 }
};

int get_flags(struct arguments *args, int argc, char *argv[]);
int check_dyndns(struct ip_addresses *ip, char *cachefile);
void update_dyndns(int s, struct arguments *args, struct ip_addresses *ip);
void update_cache(struct ip_addresses *ip, char *cachefile);
int get_local_cached_ip(struct ip_addresses *ip, char *cachefile);
int check_server_msg(int s, char *hostname);
void copy_line(char *buffer, char *string);

inline char *
  cache_file_name(struct arguments *args, char cachefile[])
{

	char user[BUFSIZE];

	bzero(cachefile, BUFSIZE);
	bzero(user, BUFSIZE);

	strncpy(user, args->user, BUFSIZE);
	snprintf(cachefile, BUFSIZE, "%s%s%s",
		"/tmp/updatedd-regfish-", args->hostname, ".ip");

	return cachefile;

}

void
  print_error(int mode, const char *msg, ...)
{

	va_list az;

	va_start(az, msg);
	if(i_syslog == 1)
		vlog_err(mode, msg, az);
	else
		vstd_err(mode, msg, az);
	va_end(az);

	return;

}

void
  dyndns(int argc, char *argv[])
{

	struct arguments args = {
		NULL, NULL, NULL, NULL,
			NULL, NULL, 0, 0, 0
	};
	struct ip_addresses ip;
	int s = 0, i;
	char *ptr, cachefile[BUFSIZE];

	(void)memset(&ip, 0, sizeof(ip));

	if(get_flags(&args, argc, argv)) {
		print_usage(argv[0], stderr);
		exit(EXIT_FAILURE);
	}

	if(args.ipv4 != NULL) {
		if(!inet_aton(args.ipv4, &ip.real_ip))
			print_error(ERROR | EXIT_FATAL, "invalid ip address: %s", args.ipv4);

		if(check_dyndns(&ip, cache_file_name(&args, cachefile)) || args.force) {
			if((ptr = get_connection(DYNDNSHOST, PORT, &s)))
				print_error(ERROR | EXIT_WARN, "%s: %s", ptr, DYNDNSHOST);
			update_dyndns(s, &args, &ip);
			if((i = check_server_msg(s, args.hostname)))
				print_error(ERROR | EXIT_FATAL, "update failed");
			else update_cache(&ip, cachefile);
			(void)close(s);
		} else print_error(INFO, "%s: update is not necessary", args.hostname);
	} else {
		if((ptr = get_connection(DYNDNSHOST, PORT, &s)))
			print_error(ERROR | EXIT_WARN, "%s: %s", ptr, DYNDNSHOST);

		update_dyndns(s, &args, &ip);
		if(check_server_msg(s, args.hostname))
			print_error(ERROR | EXIT_FATAL, "%s: update failed", args.hostname);
		(void)close(s);
	}

	return;

}

int
  get_flags(struct arguments *args, int argc, char *argv[])
{

	int c;

	while(1) {

		int option_index = 0;
		static struct option long_options[] = {
			{ "ipv4",		1, 0, '4' },
			{ "ipv6",		1, 0, '6' },
			{ "force-host",		0, 0, 'c' },
			{ "force",		0, 0, 'f' },
			{ "help",		0, 0, 'h' },
			{ "ttl",		1, 0, 'l' },
			{ "hostname",		1, 0, 'n' },
			{ "service",		1, 0, 'S' },
			{ "ismx",		0, 0, 'm' },
			{ "token",		1, 0, 't' },
			{ "version",		0, 0, 'v' },
			{ "user",		1, 0, 'u' },
			{ "with-syslog",	0, 0, 'y' },
			{ NULL,			0, 0, 0   }
		};

		c = getopt_long(argc, argv, "4:6:cfhl:n:S:mt:vu:y",
				long_options, &option_index);

		if(c == -1)
			break;

		switch(c) {
		case '4':
			args->ipv4 = optarg;
			break;
		case '6':
			args->ipv6 = optarg;
			break;
		case 'c':
			args->forcehost = 1;
			break;
		case 'f':
			args->force = 1;
			break;
		case 'l':
			if(atoi(optarg) < 300 || atoi(optarg) > 86400)
			    print_error(ERROR | EXIT_FATAL, "invalid ttl: %s", optarg);
			args->ttl = optarg;
			break;
		case 'y':
			i_syslog = 1;
			break;
		case 'h':
			print_usage(argv[0], stdout);
			exit(EXIT_SUCCESS);
		case 'm':
			args->ismx = 1;
			break;
		case 'n':
			args->hostname = optarg;
			break;
		case 'S':
			break;
		case 'v':
			(void)printf(VOUTPUT);
			exit(EXIT_SUCCESS);
		case 't':
			if(args->user != NULL)
			    print_error(ERROR | EXIT_FATAL, "use either -t or -u");
			args->token = optarg;
			break;
		case 'u':
			if(args->token != NULL)
			    print_error(ERROR | EXIT_FATAL, "use either -t or -u");

			if((args->user = (char *)malloc(strlen(optarg) + 1)) == NULL)
				print_error(ERROR | EXIT_WARN | PERR, "malloc() failed");
			strcpy(args->user, optarg);
			args->user[strlen(optarg)] = '\0';
			bzero(optarg, strlen(optarg));
			    /* prevent that anybody can see your pass with ps */
		}
	}

	if(args->hostname == NULL)
	    return 1;
	if(args->user == NULL && args->token == NULL)
	    return 1;

	return 0;

}

int
  check_dyndns(struct ip_addresses *ip, char *cachefile)
{

	if(get_local_cached_ip(ip, cachefile))
		return 1;
	if(ip->real_ip.s_addr != ip->cached.s_addr)
		return 1;
	return 0;

}

int
  get_local_cached_ip(struct ip_addresses *ip, char *cachefile)
{

	FILE *fp;
	char ip_buffer[18] = { '\0' };

	if((fp = fopen(cachefile, "r")) == NULL)
		return 1;
	if(!(fgets(ip_buffer, sizeof(ip_buffer) - 1, fp) &&
	     inet_aton(ip_buffer, &ip->cached) != 0)) {
		(void)fclose(fp);
		print_error(ERROR | EXIT_FATAL, "%s: invalid ip address", cachefile);
	}
	(void)fclose(fp);

	return 0;

}

void
  update_dyndns(int s, struct arguments *args, struct ip_addresses *ip)
{

	char buffer[4096];

	(void)memset(buffer, 0, 4096);
	if(args->ipv4) {
		strncat(buffer, "ipv4=", 4096);
		strncat(buffer, args->ipv4, 4096);
	} else {
		strncat(buffer, "thisipv4=1", 4096);
	}

	if(args->ipv6) {
		strncat(buffer, "&ipv6=", 4096);
		strncat(buffer, args->ipv6, 4096);
	}

	if(args->user) {
		int i;
		int len = strlen(args->user);

		for(i = 0; i < len; i++) {
			if(args->user[i] == ':') {
				args->user[i] = '\0';
				break;
			}
		}

		strncat(buffer, "&authtype=standard&username=", 4096);
		strncat(buffer, args->user, 4096);
		strncat(buffer, "&password=", 4096);
		strncat(buffer, args->user+i+1, 4096);
	} else if(args->token) {
		strncat(buffer, "&authtype=secure&token=", 4096);
		strncat(buffer, args->token, 4096);
	} else
	    print_error(ERROR | EXIT_FATAL, "login is missing");

	if(args->ttl) {
		strncat(buffer, "&ttl=", 4096);
		strncat(buffer, args->ttl, 4096);
	}

	print_error(DEBUGINFO,
		"GET /dyndns/2/?%s&fqdn=%s&forcehost=%i&ismx=%i HTTP/1.1\r\n"
		"Host: %s\r\n"
		"Connection: close\r\n"
		"Pragma: no-cache\r\n\r\n",
		buffer, args->hostname, args->forcehost, args->ismx, DYNDNSHOST);


	(void)dprintf(s,
		"GET /dyndns/2/?%s&fqdn=%s&forcehost=%i&ismx=%i HTTP/1.1\r\n"
		"Host: %s\r\n"
		"Connection: close\r\n"
		"Pragma: no-cache\r\n\r\n",
		buffer, args->hostname, args->forcehost, args->ismx, DYNDNSHOST);

	return;

}

void
  update_cache(struct ip_addresses *ip, char *cachefile)
{

	FILE *fp;

	(void)unlink(cachefile);
	if((fp = fopen(cachefile, "a")) == NULL)
		print_error(ERROR | EXIT_FATAL | PERR, "fopen() failed");
	(void)fprintf(fp, "%s", inet_ntoa(ip->real_ip));
	fclose(fp);

	return;

}

int
  check_server_msg(int s, char *hostname)
{

	int n;
	char server_msg[MAXDATA], *ptr;

	/* get server_msg */
	(void)memset(server_msg, 0, sizeof(server_msg));
	if(read(s, server_msg, sizeof(server_msg) - 1) < 0)
		print_error(ERROR | EXIT_WARN | PERR, "read() failed");

	print_error(DEBUGINFO, "\n\nServer message:"
		     "\n--------------------------------------\n"
		     "%s--------------------------------------\n\n",
		     server_msg);

	if(strstr(server_msg, "HTTP/1.1 200 OK") ||
	   strstr(server_msg, "HTTP/1.0 200 OK")) {

		(void)strtok(server_msg, "\n");
		while((ptr = strtok(NULL, "\n")) != NULL) {
			for(n=0; return_codes[n].code != NULL; n++) {
				char buf[256];
				(void)memset(buf, 0, 256);
				snprintf(buf, 256, "|%s|", return_codes[n].code);
				if(strstr(ptr, buf)) {
					if(return_codes[n].error == 1) {
						print_error(WARNING, "%s: %s",
						    hostname, return_codes[n].message);
						return 1;
					} else {
						print_error(NOTICE, "%s: %s",
						    hostname, return_codes[n].message);
					}
				}
			}
		}
	} else {
		print_error(ERROR | EXIT_FATAL, "Internal Server Error");
	}

	return 0;
}
